#!/usr/bin/bash
# shellcheck shell=bash

AASSU__LOG_LEVEL=1
AASSU__DEFAULT_NOOP=0
AASSU__DEFAULT_RELEASE_VARIANT=4
AASSU__MAX_VERSION_CODE=1073741808

show_help() {

    cat <<'HELP_EOF'
android-app-semver-spec-utils.sh provides utils to convert versions
that are compliant with the `android-app-semver` spec.


Available commands:
    name-to-code     Convert a version name to a version code.
    code-to-name     Convert a version code to a version name.


Usage:
  android-app-semver-spec-utils.sh [command_options] name-to-code \
    <version_name> [[release_variant] [noop]]
  android-app-semver-spec-utils.sh [command_options] code-to-name \
    <version_code>


Available command_options:
  [ -h | --help ]    Display this help screen.
  [ -v ]             Set log level to `DEBUG`.



For the `name-to-code` command:
- The `release_variant` and `noop` arguments are optional. Either
  both should be passed, or none, or only `release_variant`.
  The default value is `4` for `release_variant` and `0` for `noop`.
- The `version_name` must be in the format
  `major.minor.patch(-(alpha|beta|rc).<revision>)`.
  - The `major` value must be between `0-63` (inclusive).
  - The `minor` value must be between `0-511` (inclusive).
  - The `patch` value must be between `0-31` (inclusive).
  - The `(-(alpha|beta|rc).<revision>)` part is optional and if not
    passed, it is conidered a `stable` `release_type`. If passed, then
    `revision` must be between `1-15` (inclusive), like `-beta.1`.
- The `release_variant` value must be between `0-7` (inclusive).
- The `noop` value must be between `0-1` (inclusive).

Examples:
android-app-semver-spec-utils.sh name-to-code 0.1.0
android-app-semver-spec-utils.sh name-to-code 0.1.0-beta.1 0
android-app-semver-spec-utils.sh name-to-code 63.511.31 7 1



For the `code-to-name` command:
- The `version_code` value must be between `0-1073741808` (inclusive)
  and must be a valid value as per spec.

Examples:
android-app-semver-spec-utils.sh code-to-name 268437552



Pass the `-v` flag to print additonal info for version parts, which
is useful to debug issues.
HELP_EOF

}

main() {

    local return_value

    if [ "$1" = "-h" ] || [ "$1" = "--help" ]; then
        show_help || return $?
        return 0
    else
        if [ "${1:-}" = "-v" ]; then
            AASSU__LOG_LEVEL=2
            shift
        fi

        if [ "$1" = "name-to-code" ]; then
            shift
            name_to_code "$@"
            return_value=$?
        elif [ "$1" = "code-to-name" ]; then
            shift
            code_to_name "$@"
            return_value=$?
        else
            echo "Invalid command '$1'." 1>&2
            return_value=64
        fi
        if [ $return_value -eq 64 ]; then # EX__USAGE
            echo ""
            show_help
        fi
        return $return_value
    fi

}

name_to_code() {

    if [ $# -lt 1 ] || [ $# -gt 3 ]; then
        echo "Invalid argument count $#. The \"name_to_code\" method expects 1 to 3 arguments." 1>&2
        printf 'Arguments: %s\n' "$*" 1>&2
        return 64
    fi

    local version_name="${1/v/}"
    local release_variant="${2:-$AASSU__DEFAULT_RELEASE_VARIANT}"
    local noop="${3:-$AASSU__DEFAULT_NOOP}"

    local version_name_regex='^([0-9]|[1-5][0-9]|6[0-3])\.([0-9]|[1-9][0-9]|[1-4][0-9][0-9]|50[0-9]|51[0-1])\.([0-9]|[1-2][0-9]|3[0-1])(-(alpha|beta|rc)\.([1-9]|1[0-5]))?$'

    if ! printf "%s" "$version_name" | grep -qE "$version_name_regex"; then
        echo "The version_name '$version_name' is not valid as per android-app-semver spec compliant with semantic version '2.0.0' and  in the format 'major.minor.patch(-(alpha|beta|rc).<revision>)'. https://semver.org/spec/v2.0.0.html." 1>&2
        return 64
    fi

    if [[ ! "$release_variant" =~ ^[0-7]$ ]]; then
        echo "The release_variant '$release_variant' for version_name '$version_name' is not valid and must be between 0-7 (inclusive)." 1>&2
        return 64
    fi

    if [[ ! "$noop" =~ ^[0-1]$ ]]; then
        echo "The noop '$noop' for version_name '$version_name' is not valid and must be between 0-1 (inclusive)." 1>&2
        return 64
    fi


    local major
    local minor
    local patch
    local release_type
    local release_type_name
    local revision
    local version_code
    local version_code_binary

    major="$(printf "%s" "$version_name" | sed -E "s/$version_name_regex/\1/")" || return $?
    minor="$(printf "%s" "$version_name" | sed -E "s/$version_name_regex/\2/")" || return $?
    patch="$(printf "%s" "$version_name" | sed -E "s/$version_name_regex/\3/")" || return $?
    release_type_name="$(printf "%s" "$version_name" | sed -E "s/$version_name_regex/\5/")" || return $?
    revision="$(printf "%s" "$version_name" | sed -E "s/$version_name_regex/\6/")" || return $?

    case "$release_type_name" in
        alpha) release_type=0;;
        beta) release_type=1;;
        rc) release_type=2;;
        *) release_type_name="stable"; release_type=3; revision=0;;
    esac

    if [[ "$AASSU__LOG_LEVEL" == "2" ]]; then
        print_verbose_version_info || return $?
    fi

    if [ $release_type -lt 3 ] && [ $revision -lt 1 ]; then
        echo "The revision '$revision' for version_name '$version_name' is not valid and must not be 0 for a non-stable release with type '$release_type_name'." 1>&2
        return 1
    fi


    version_code=0
    version_code=$(( version_code + revision )) || return $?                         # 0 000 000000 000000000 00000 00 1111
    version_code=$(( version_code + (release_type << 4) )) || return $?              # 0 000 000000 000000000 00000 11 0000
    version_code=$(( version_code + (patch << 6) )) || return $?                     # 0 000 000000 000000000 11111 00 0000
    version_code=$(( version_code + (minor << 11) )) || return $?                    # 0 000 000000 111111111 00000 00 0000
    version_code=$(( version_code + (major << 20) )) || return $?                    # 0 000 111111 000000000 00000 00 0000
    version_code=$(( version_code + (release_variant << 26) )) || return $?          # 0 111 000000 000000000 00000 00 0000
    version_code=$(( version_code + (noop << 29) )) || return $?                     # 1 000 000000 000000000 00000 00 0000

    if [ $version_code -lt 0 ] || [ $version_code -gt $AASSU__MAX_VERSION_CODE ]; then
        echo "The version_code '$version_code' for version_name '$version_name' is not valid and must be between 0-$AASSU__MAX_VERSION_CODE (inclusive)." 1>&2
        return 1
    fi


    version_code_binary="$(decimal_to_binary "$version_code" 30)" || return $?
    version_code_binary="$(printf "%s\n" "$version_code_binary" | sed -E 's/(.{1})(.{3})(.{6})(.{9})(.{5})(.{2})(.{4})/\1 \2 \3 \4 \5 \6 \7/')" || return $?


    print_formatted_version_info || return $?

}


code_to_name() {

    if [ $# -ne 1 ]; then
        echo "Invalid argument count $#. The \"name_to_code\" method expects 1 argument." 1>&2
        printf 'Arguments: %s\n' "$*" 1>&2
        return 64
    fi

    local version_code="$1"

    if [[ ! "$version_code" =~ ^[0-9]+$ ]] || [ "$version_code" -gt $AASSU__MAX_VERSION_CODE ]; then
        echo "The version_code '$version_code' is not valid and must be between 0-$AASSU__MAX_VERSION_CODE (inclusive)." 1>&2
        return 64
    fi

    version_code_binary="$(decimal_to_binary "$version_code" 30)" || return $?
    version_code_binary="$(printf "%s\n" "$version_code_binary" | sed -E 's/(.{1})(.{3})(.{6})(.{9})(.{5})(.{2})(.{4})/\1 \2 \3 \4 \5 \6 \7/')" || return $?

    if [[ "$AASSU__LOG_LEVEL" == "2" ]]; then
        echo "version_code_binary: $version_code_binary"
    fi


    local major
    local minor
    local patch
    local release_type
    local release_type_name
    local revision
    local version_name

    noop=$(( (version_code & 536870912) >> 29 )) || return $?                    # 1 000 000000 000000000 00000 00 0000
    release_variant=$(( (version_code & 469762048) >> 26 )) || return $?         # 0 111 000000 000000000 00000 00 0000
    major=$(( (version_code & 66060288) >> 20 )) || return $?                    # 0 000 111111 000000000 00000 00 0000
    minor=$(( (version_code & 1046528) >> 11 )) || return $?                     # 0 000 000000 111111111 00000 00 0000
    patch=$(( (version_code & 1984) >> 6 )) || return $?                         # 0 000 000000 000000000 11111 00 0000
    release_type=$(( (version_code & 48) >> 4 )) || return $?                    # 0 000 000000 000000000 00000 11 0000
    revision=$(( (version_code & 15) )) || return $?                             # 0 000 000000 000000000 00000 00 1111


    version_name="$major.$minor.$patch"
    case "$release_type" in
        0) release_type_name=alpha;;
        1) release_type_name=beta;;
        2) release_type_name=rc;;
        *) release_type_name=""
    esac

    if [[ "$AASSU__LOG_LEVEL" == "2" ]]; then
        print_verbose_version_info || return $?
    fi


    if [ -n "$release_type_name" ]; then
        if [ $revision -lt 1 ]; then
            echo "The revision '$revision' for version_code '$version_code' is not valid and must not be 0 for a non-stable release with type '$release_type_name'." 1>&2
            return 1
        fi

        version_name+="-$release_type_name.$revision"
    else
        if [ "$revision" -ne 0 ]; then
            echo "The revision '$revision' for version_code '$version_code' is not valid and must be 0 for a stable release." 1>&2
            return 64
        fi
    fi


    print_formatted_version_info || return $?

}

print_verbose_version_info() {

    echo "noop: $noop"
    echo "release_variant: $release_variant"
    echo "major: $major"
    echo "minor: $minor"
    echo "patch: $patch"
    echo "release_type: $release_type (${release_type_name:-stable})"
    echo "revision: $revision"
    echo

}

print_formatted_version_info() {

    printf "version_name: %17s, version_code: %10s (%s), release_variant: %1s" "$version_name" "$version_code" "$version_code_binary" "$release_variant" || return $?
    if [ "$noop" -gt 0 ]; then
        printf ", noop: %1s" "$noop" || return $?
    fi
    echo

}

decimal_to_binary() {

    local decimal="$1"
    local min_digits="$2"
    if [[ ! "$min_digits" =~ ^[0-9]+$ ]]; then
        min_digits=8
    fi

    local binary
    local binary_digits
    local zeros=""
    local zeros_required

    binary="$(printf "%s\n" "ibase=10;obase=2;$decimal" | bc)" || return $?

    binary_digits=${#binary}
    if [ "$binary_digits" -lt $min_digits ]; then
        zeros_required=$((min_digits - binary_digits))
        zeros=""
        for (( i = 0; i < "$zeros_required"; i++ )); do
            zeros="${zeros}0"
        done
    fi

    printf "%s\n" "$zeros$binary"

}



# If running in bash, run script logic, otherwise exit with usage error
if [ -n "${BASH_VERSION:-}" ]; then
    # If script is sourced, return with success, otherwise call main function
    # - https://stackoverflow.com/a/28776166/14686958
    # - https://stackoverflow.com/a/29835459/14686958
    if (return 0 2>/dev/null); then
        return 0 # EX__SUCCESS
    else
        main "$@"
        exit $?
    fi
else
    (echo "${0##*/} must be run with the bash shell."; exit 64) # EX__USAGE
fi
